/**
 * Hostkey - public key crypto. This is the testcase!
 * @author Christian Grothoff
 * @file test/hostkeytest.c
 **/

#include "config.h"
#include <stdio.h>
#include <string.h>
#include <gcrypt.h>
#include "util/hashing.h"
#include <netinet/in.h>
#include "util/logging.h"
#include "util/xmalloc.h"
#include "util/hostkey.h"
#include "util/returnvalues.h"

#define TESTSTRING "Hello World\0"
#define MAX 20
#define ITER 100

static int testEncryptDecrypt() {
  HOSTKEY hostkey;
  PublicKey pkey;
  RSAEncryptedData target;
  char result[MAX];
  int i;
  time_t start;


  hostkey = makeHostkey();
  getPublicKey(hostkey, &pkey);

  time(&start);
  for (i=0;i<ITER;i++) {
    fprintf(stderr, ".");
    if (SYSERR == encryptHostkey(TESTSTRING,strlen(TESTSTRING)+1,&pkey,
				 &target))
      return SYSERR;
    if (-1 == decryptHostkey(hostkey, &target, result, MAX))
      return SYSERR;
    if (strncmp(TESTSTRING, result,
		strlen(TESTSTRING)) != 0) {
      printf("%s != %.*s - testEncryptDecrypt failed!\n",
	     TESTSTRING, MAX, result);
      return SYSERR;
    }
  }
  printf("%d RSA encrypt/decrypt operations %ds\n", 
	 ITER,
	 (int) (time(NULL)-start));
  freeHostkey(hostkey);
  return OK;
}

static int testSignVerify() {
  HOSTKEY hostkey;
  Signature sig;
  PublicKey pkey;
  int i;
  time_t start;
  
  hostkey = makeHostkey();
  getPublicKey(hostkey, &pkey);
  time(&start);
  for (i=0;i<ITER;i++) {
    fprintf(stderr, ".");
    if (SYSERR == sign(hostkey, strlen(TESTSTRING), TESTSTRING, &sig)) {
      return SYSERR;
    }
    if (SYSERR == verifySig(TESTSTRING, strlen(TESTSTRING), &sig, &pkey)) {
      printf("testSignVerify failed!\n");
      return SYSERR;
    }
  }
  printf("%d RSA sign/verify operations %ds\n", 
	 ITER, 
	 (int) (time(NULL)-start));
  freeHostkey(hostkey);
  return OK;
}

static int testHostkeyEncoding() {
  HOSTKEY hostkey;
  HostKeyEncoded * encoding;
  PublicKey pkey;
  RSAEncryptedData target;
  char result[MAX];
  int i;
  time_t start;

  hostkey = makeHostkey();

  time(&start);
  for (i=0;i<ITER;i++) {
    fprintf(stderr, ".");
    getPublicKey(hostkey, &pkey);
    if (SYSERR == encryptHostkey(TESTSTRING,
				 strlen(TESTSTRING)+1,
				 &pkey,
				 &target))
      return SYSERR;
    encoding = encodeHostkey(hostkey);
    freeHostkey(hostkey);
    if (encoding == NULL)
      return SYSERR;
    hostkey = decodeHostkey(encoding);
    free(encoding);
    if (-1 == decryptHostkey(hostkey, &target, result, MAX)) {
      return SYSERR;
    }  
    if (strncmp(TESTSTRING, result,
		strlen(TESTSTRING)) != 0) {
      printf("%s != %.*s - testEncryptDecrypt failed!\n",
	     TESTSTRING, 
	     (int) strlen(TESTSTRING), 
	     result);
      return SYSERR;
    }
  }  
  freeHostkey(hostkey);
  printf("%d RSA encrypt/encode/decode/decrypt operations %ds\n", 
	 ITER, 
	 (int) (time(NULL)-start));
  return OK;
}


int main(int argc, char * argv[]) {
  int failureCount = 0;

  initRAND();
#ifndef USE_OPENSSL
  gcry_control (GCRYCTL_INITIALIZATION_FINISHED, 0);
  /* No valuable keys are create, so we can speed up our RNG. */
  gcry_control (GCRYCTL_ENABLE_QUICK_RANDOM, 0);
#endif
  
  if (OK != testEncryptDecrypt())
     failureCount++;
  if (OK != testSignVerify())
    failureCount++;     
  if (OK != testHostkeyEncoding()) 
    failureCount++;


  if (failureCount == 0)
    return 0;
  else {
    printf("\n\n%d TESTS FAILED!\n\n",failureCount);
    return -1;
  }
} // end of main
