/*
     This file is part of GNUnet.
     (C) 2001, 2002, 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Hostkey - public key cryptography routines.
 *
 * Todo:
 * - match formats of OpenSSL & libgcrypt (possible?)
 *
 * @author Christian Grothoff
 * @file util/hostkey.c
 **/

#include "config.h"
#include "util/hostkey.h"
#include "util/xmalloc.h"
#include "util/returnvalues.h"
#include "util/logging.h"
#include "util/cron.h"
#include "util/hashing.h"
#include <netinet/in.h>


#define HOSTKEY_LEN 2048
#define EXTRA_CHECKS YES

/**
 * Initialize Random number generator.
 **/
void initRAND() {
#ifndef USE_OPENSSL
  gcry_control(GCRYCTL_DISABLE_SECMEM, 0);
  if (! gcry_check_version(GCRYPT_VERSION))
    errexit("FATAL: libgcrypt has not the exptected version %s\n",
	    GCRYPT_VERSION);
#endif
  srand((unsigned int)time(NULL));
}

/**
 * This HostKey implementation uses RSA.
 **/
HOSTKEY makeHostkey() {
#ifdef USE_OPENSSL
  return RSA_generate_key(HOSTKEY_LEN, 65535, NULL, 0);
#else
  HOSTKEY s_key;
  GCRY_SEXP s_keyparam;
  int rc;

  rc = gcry_sexp_build(&s_keyparam, 
		       NULL, 
		       "(genkey(rsa(nbits %d)(rsa-use-e 3:257)))",
		       HOSTKEY_LEN);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: failed to convert keyparam: %s\n", 
	gcry_strerror (rc));
    return NULL;
  }
  rc = gcry_pk_genkey(&s_key,
		      s_keyparam);
  gcry_sexp_release(s_keyparam);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: key generation failed: %s\n", 
	gcry_strerror (rc));
    return NULL;
  }

#if EXTRA_CHECKS
  if (gcry_pk_testkey(s_key)) {
    LOG(LOG_ERROR,
	"ERROR: freshly created hostkey is not sane!\n");
    return NULL;
  }
#endif
  return s_key;
#endif
}

/**
 * Free memory occupied by hostkey
 **/
void freeHostkey(HOSTKEY hostkey) {
#ifdef USE_OPENSSL
  RSA_free(hostkey);
#else
  gcry_sexp_release(hostkey);
#endif
}

#ifndef USE_OPENSSL
static int key_from_sexp( GCRY_MPI *array,
			  GCRY_SEXP sexp, 
			  const char *topname, 
			  const char *elems ) {
  GCRY_SEXP list, l2;
  const char *s;
  int i, idx;
  
  list = gcry_sexp_find_token( sexp, topname, 0 );
  if( !list )
    return GCRYERR_INV_OBJ;
  l2 = gcry_sexp_cadr( list );
  gcry_sexp_release ( list );
  list = l2;
  if( !list )
    return GCRYERR_NO_OBJ;
  
  idx = 0;
  for(s=elems; *s; s++, idx++ ) {
    l2 = gcry_sexp_find_token( list, s, 1 );
    if( !l2 ) {
      for(i=0; i<idx; i++) {
	gcry_free( array[i] );
	array[i] = NULL;
      }
      gcry_sexp_release ( list );
      return GCRYERR_NO_OBJ; /* required parameter not found */
    }
    array[idx] = gcry_sexp_nth_mpi( l2, 1, GCRYMPI_FMT_USG );
    gcry_sexp_release ( l2 );
    if( !array[idx] ) {
      for(i=0; i<idx; i++) {
	gcry_free( array[i] );
	array[i] = NULL;
      }
      gcry_sexp_release ( list );
      return GCRYERR_INV_OBJ; /* required parameter is invalid */
    }
  }
  gcry_sexp_release ( list );  
  return 0;
}
#endif

/**
 * Extract the public key of the host.
 * @param hostkey the hostkey to extract into the result.
 * @param result where to write the result.
 **/
void getPublicKey(HOSTKEY hostkey,
		  PublicKey * result) {
#ifdef USE_OPENSSL
  unsigned short sizen;
  unsigned short sizee;
  unsigned short size;

  sizen = BN_num_bytes(hostkey->n);
  sizee = BN_num_bytes(hostkey->e);
  size = sizen + sizee+2*sizeof(unsigned short);
  if (size != sizeof(PublicKey)-sizeof(result->padding)) 
    errexit("FATAL: sizeof public key does not match size (%d!=%d)\n",
	    size, 
	    sizeof(PublicKey)-sizeof(result->padding));
  if (RSA_KEY_LEN != sizen+sizee) 
    errexit("FATAL: PublicKey datastructure wrong (%d+%d!=%d)!\n",
	    sizen, 
	    sizee, 
	    RSA_KEY_LEN);
  result->len = htons(size);
  result->sizen = htons(sizen);
  result->padding = 0;
  BN_bn2bin(hostkey->n,
	    &result->key[0]);
  BN_bn2bin(hostkey->e,
	    &result->key[sizen]);
#else
  GcryMPI skey[2];
  int size;
  int rc;
  
  rc = key_from_sexp(skey, 
		     hostkey, 
		     "public-key", 
		     "ne");
  if (rc)
    rc = key_from_sexp(skey, 
		       hostkey, 
		       "private-key", 
		       "ne");    
  if (rc)
    rc = key_from_sexp(skey, 
		       hostkey, 
		       "rsa", 
		       "ne");    
  if (rc) 
    errexit("FATAL: getPublicKey: key_from_sexp failed: %s\n", 
	    gcry_strerror(rc));
  
  result->len = htons(sizeof(PublicKey) - sizeof(result->padding));
  result->sizen = htons(RSA_ENC_LEN);
  result->padding = 0;
  size = RSA_ENC_LEN;
  rc = gcry_mpi_print(GCRYMPI_FMT_USG, 
		      &result->key[0], 
		      &size, 
		      skey[0]);
  if (rc) 
    errexit("FATAL: gcry_mpi_print of n failed: %s\n",
	    gcry_strerror(rc));
  if (size != RSA_ENC_LEN)
    errexit("FATAL: gcry_mpi_print assert (n) failed! (%d != %d)\n",
	    size,
	    RSA_ENC_LEN);

  size = RSA_KEY_LEN - RSA_ENC_LEN; 
  rc = gcry_mpi_print(GCRYMPI_FMT_USG, 
		      &result->key[RSA_ENC_LEN], 
		      &size, 
		      skey[1]);
  if (rc) 
    errexit("FATAL: gcry_mpi_print of e failed: %s\n",
	    gcry_strerror(rc));
  if (size != RSA_KEY_LEN - RSA_ENC_LEN)
    errexit("FATAL: gcry_mpi_print assert (e) failed! (%d != %d)\n",
	    size, RSA_KEY_LEN - RSA_ENC_LEN);
#endif
}


/**
 * Internal: publicKey => RSA-Key
 **/
static HOSTKEY public2Hostkey(PublicKey * publicKey) {
#ifdef USE_OPENSSL
  HOSTKEY result;
  int sizen;
  int sizee;

  result = RSA_new();
  if (ntohs(publicKey->len) != sizeof(PublicKey)-sizeof(publicKey->padding)) {
    LOG(LOG_ERROR,
	"ERROR: public2Hostkey: received invalid publicKey (size=%d)\n",
	ntohs(publicKey->len));
    return NULL;
  }
  sizen = ntohs(publicKey->sizen);
  sizee = ntohs(publicKey->len) - sizen - 2*sizeof(unsigned short);
  if ( (sizen != RSA_ENC_LEN) || 
       (sizee + sizen != RSA_KEY_LEN)) {
    LOG(LOG_ERROR,
	"ERROR: public2Hostkey: received invalid publicKey (sizee=%d, sizen=%d)\n",
	sizee,
	sizen);
    return NULL;
  }
  result->n = BN_bin2bn(&publicKey->key[0], 
			sizen, 
			NULL);
  result->e = BN_bin2bn(&publicKey->key[sizen],
			sizee, 
			NULL);
  return result;
#else
  HOSTKEY result;
  GcryMPI n;
  GcryMPI e;
  size_t size;
  size_t erroff;
  int rc;
  // len = sizen + sizee + 4 == publickey - padding!
  if ( ( ntohs(publicKey->sizen) != RSA_ENC_LEN ) ||
       ( ntohs(publicKey->len) != sizeof(PublicKey) - sizeof(publicKey->padding)) ) {
    LOG(LOG_ERROR,
	"ERROR: public2Hostkey: received invalid publicKey (%d, %d)\n",
	ntohs(publicKey->len),
	ntohs(publicKey->sizen)); 
    return NULL;
  }
  size = RSA_ENC_LEN;
  rc = gcry_mpi_scan(&n,
		     GCRYMPI_FMT_USG,
		     &publicKey->key[0],
		     &size);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: gcry_mpi_scan of n failed (%s)\n",
	gcry_strerror(rc));
    return NULL;
  }
  size = RSA_KEY_LEN - RSA_ENC_LEN;
  rc = gcry_mpi_scan(&e,
		     GCRYMPI_FMT_USG,
		     &publicKey->key[RSA_ENC_LEN],
		     &size);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: gcry_mpi_scan of e failed (%s)\n",
	gcry_strerror(rc));
    gcry_mpi_release(n);
    return NULL;
  }
  rc = gcry_sexp_build(&result, 
		       &erroff,
 		       "(public-key(rsa(n %m)(e %m)))",
		       n,
		       e);
  gcry_mpi_release(n);
  gcry_mpi_release(e);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: gcry_sexp_build of public key failed (%s at position %d)\n",
	gcry_strerror(rc),
	erroff);
    return NULL;
  }  
  return result;
#endif
}

/**
 * Encode the private key in a format suitable for
 * storing it into a file.
 * @returns encoding of the private key.
 *    The first 4 bytes give the size of the array, as usual.
 **/
HostKeyEncoded * encodeHostkey(HOSTKEY hostkey) {
#ifdef USE_OPENSSL
  /*
               BIGNUM *n;               public modulus
               BIGNUM *e;               public exponent
               BIGNUM *d;               private exponent
               BIGNUM *p;               secret prime factor
               BIGNUM *q;               secret prime factor
               BIGNUM *dmp1;            d mod (p-1)
               BIGNUM *dmq1;            d mod (q-1)
               BIGNUM *iqmp;            q^-1 mod p
  */
  unsigned short sizen;
  unsigned short sizee;
  unsigned short sized;
  unsigned short sizep;
  unsigned short sizeq;
  unsigned short sizedmp1;
  unsigned short sizedmq1;
  unsigned short sizeiqmp;
  unsigned short size;
  HostKeyEncoded * retval;

  sizen = BN_num_bytes(hostkey->n);
  sizee = BN_num_bytes(hostkey->e);
  sized = BN_num_bytes(hostkey->d);
  if (hostkey->p != NULL)
    sizep = BN_num_bytes(hostkey->p);
  else
    sizep = 0;
  if (hostkey->q != NULL)
    sizeq = BN_num_bytes(hostkey->q);
  else
    sizeq = 0;
  if (hostkey->dmp1 != NULL)
    sizedmp1 = BN_num_bytes(hostkey->dmp1);
  else
    sizedmp1 = 0;
  if (hostkey->dmq1 != NULL)
    sizedmq1 = BN_num_bytes(hostkey->dmq1);
  else
    sizedmq1 = 0;
  if (hostkey->iqmp != NULL)
    sizeiqmp = BN_num_bytes(hostkey->iqmp);
  else
    sizeiqmp = 0;
  size = sizen+sizee+sized+sizep+sizeq+sizedmp1+sizedmq1+sizeiqmp+sizeof(HostKeyEncoded);
  retval = (HostKeyEncoded *) xmalloc(size,
				      "encodeHostkey: hostkey");
  retval->len = htons(size);
  retval->sizen = htons(sizen);
  retval->sizee = htons(sizee);
  retval->sized = htons(sized);
  retval->sizep = htons(sizep);
  retval->sizeq = htons(sizeq);
  retval->sizedmp1 = htons(sizedmp1);
  retval->sizedmq1 = htons(sizedmq1);
  BN_bn2bin(hostkey->n, &retval->key[0]);
  BN_bn2bin(hostkey->e, &retval->key[0+sizen]);
  BN_bn2bin(hostkey->d, &retval->key[0+sizen+sizee]);
  if (hostkey->p != NULL)
    BN_bn2bin(hostkey->p, 
	      &retval->key[0+sizen+sizee+sized]);
  if (hostkey->q != NULL)
    BN_bn2bin(hostkey->q, 
	      &retval->key[0+sizen+sizee+sized+sizep]);
  if (hostkey->dmp1 != NULL)
    BN_bn2bin(hostkey->dmp1, 
	      &retval->key[0+sizen+sizee+sized+sizep+sizeq]);
  if (hostkey->dmq1 != NULL)
    BN_bn2bin(hostkey->dmq1, 
	      &retval->key[0+sizen+sizee+sized+sizep+sizeq+sizedmp1]);
  if (hostkey->iqmp != NULL)
    BN_bn2bin(hostkey->iqmp, 
	      &retval->key[0+sizen+sizee+sized+sizep+sizeq+sizedmp1+sizedmq1]);
  return retval;
#else
  /* libgcrypt */

  HostKeyEncoded * retval;
  GcryMPI pkv[6];
  void * pbu[6];
  size_t sizes[6];
  int rc;
  int i;
  int size;

#if EXTRA_CHECKS
  if (gcry_pk_testkey(hostkey)) {
    LOG(LOG_ERROR,
	"ERROR: encodeHostkey: hostkey is not sane!\n");
    return NULL;
  }
#endif

  memset(pkv, 0, sizeof(GcryMPI) * 6);
  rc = key_from_sexp(pkv,
		     hostkey,
		     "private-key",
		     "nedpqu");
  if (rc)
    rc = key_from_sexp(pkv,
		       hostkey,
		       "rsa",
		       "nedpqu");
  if (rc) 
    rc = key_from_sexp(pkv,
		       hostkey,
		       "private-key",
		       "nedpq");
  if (rc)
    rc = key_from_sexp(pkv,
		       hostkey,
		       "rsa",
		       "nedpq");
  if (rc) 
    rc = key_from_sexp(pkv,
		       hostkey,
		       "private-key",
		       "ned");
  if (rc)
    rc = key_from_sexp(pkv,
		       hostkey,
		       "rsa",
		       "ned");
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: encodeHostkey: key_from_sexp failed: %s\n", 
	gcry_strerror (rc));
    return NULL;
  }
  size = sizeof(HostKeyEncoded);
  for (i=0;i<6;i++) {
    if (pkv[i] != NULL) {
      rc = gcry_mpi_aprint(GCRYMPI_FMT_USG,
			   &pbu[i],
			   &sizes[i],
			   pkv[i]);
      size += sizes[i];
      if (rc) {
	LOG(LOG_ERROR,
	    "ERROR: gcry_mpi_aprint failed: %s\n", 
	    gcry_strerror (rc));
	while (i>0) {
	  freeNonNull(pbu[--i], 
		      "key conversion buffer");
	}
	for (i=0;i<6;i++)
	  if (pkv[i] != NULL)
	    gcry_mpi_release(pkv[i]);
	return NULL;
      }
    } else {
      pbu[i] = NULL;
      sizes[i] = 0;
    }
  }
  if (size >= 65536) 
    errexit("FATAL: size of serialized private key >= 64k\n");  
  retval = xmalloc(size,
		   "allocting host key encoded");
  retval->len = htons(size);
  i = 0;
  retval->sizen = htons(sizes[0]);
  memcpy(&retval->key[i], 
	 pbu[0],
	 sizes[0]);
  i += sizes[0];
  retval->sizee = htons(sizes[1]);
  memcpy(&retval->key[i], 
	 pbu[1],
	 sizes[1]);
  i += sizes[1];
  retval->sized = htons(sizes[2]);
  memcpy(&retval->key[i], 
	 pbu[2],
	 sizes[2]);
  i += sizes[2];
  /* swap p and q! */
  retval->sizep = htons(sizes[4]);
  memcpy(&retval->key[i], 
	 pbu[4],
	 sizes[4]);
  i += sizes[4];
  retval->sizeq = htons(sizes[3]);
  memcpy(&retval->key[i], 
	 pbu[3],
	 sizes[3]);
  i += sizes[3];
  retval->sizedmp1 = htons(0);
  retval->sizedmq1 = htons(0);
  memcpy(&retval->key[i], 
	 pbu[5],
	 sizes[5]);
  for (i=0;i<6;i++) {
    if (pkv[i] != NULL)
      gcry_mpi_release(pkv[i]);
    freeNonNull(pbu[i], 
		"freeing temporaries in hostkey serialization");
  }
  return retval;
#endif
}

/**
 * Decode the private key from the file-format back
 * to the "normal", internal format.
 **/
HOSTKEY decodeHostkey(HostKeyEncoded * encoding) {
#ifdef USE_OPENSSL
  unsigned short sizen;
  unsigned short sizee;
  unsigned short sized;
  unsigned short sizep;
  unsigned short sizeq;
  unsigned short sizedmp1;
  unsigned short sizedmq1;
  unsigned short size;
  unsigned short sum;

  HOSTKEY result = (HOSTKEY) RSA_new();
  size    = ntohs(encoding->len) - sizeof(HostKeyEncoded);
  sizen   = ntohs(encoding->sizen);
  sizee   = ntohs(encoding->sizee);
  sized   = ntohs(encoding->sized);
  sizep   = ntohs(encoding->sizep);
  sizeq   = ntohs(encoding->sizeq);
  sizedmp1= ntohs(encoding->sizedmp1);
  sizedmq1= ntohs(encoding->sizedmq1);
  sum = 0;
  result->n= BN_bin2bn(&encoding->key[sum], sizen, NULL); sum += sizen;
  result->e= BN_bin2bn(&encoding->key[sum], sizee, NULL); sum += sizee;
  result->d= BN_bin2bn(&encoding->key[sum], sized, NULL); sum += sized;
  if (sizep != 0) {
    result->p = BN_bin2bn(&encoding->key[sum], sizep, NULL); sum += sizep;
  } else
    result->p = NULL;
  if (sizeq != 0) {
    result->q = BN_bin2bn(&encoding->key[sum], sizeq, NULL); sum += sizeq;
  } else
    result->q = NULL;
  if (sizedmp1 != 0) {
    result->dmp1= BN_bin2bn(&encoding->key[sum], sizedmp1, NULL); sum += sizedmp1;
  } else
    result->dmp1 = NULL;
  if (sizedmq1 != 0) {
    result->dmq1 = BN_bin2bn(&encoding->key[sum], sizedmq1, NULL); sum += sizedmq1;
  } else
    result->dmq1 = NULL;
  if (size - sum > 0) 
    result->iqmp= BN_bin2bn(&encoding->key[sum], size-sum, NULL);
  else
    result->iqmp = NULL;
  return result;
#else
  HOSTKEY res;
  GcryMPI n,e,d,p,q,u;
  int rc;
  size_t size;
  int pos;

  pos = 0;
  size = ntohs(encoding->sizen);
  rc = gcry_mpi_scan(&n,
		     GCRYMPI_FMT_USG,
		     &encoding->key[pos],
		     &size);
  pos += ntohs(encoding->sizen);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: could not decode hostkey (%s)\n",
	gcry_strerror(rc));
    return NULL;
  }
  size = ntohs(encoding->sizee);
  rc = gcry_mpi_scan(&e,
		     GCRYMPI_FMT_USG,
		     &encoding->key[pos],
		     &size);
  pos += ntohs(encoding->sizee);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: could not decode hostkey (%s)\n",
	gcry_strerror(rc));
    gcry_mpi_release(n);
    return NULL;
  }
  size = ntohs(encoding->sized);
  rc = gcry_mpi_scan(&d,
		     GCRYMPI_FMT_USG,
		     &encoding->key[pos],
		     &size);
  pos += ntohs(encoding->sized);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: could not decode hostkey (%s)\n",
	gcry_strerror(rc));
    gcry_mpi_release(n);
    gcry_mpi_release(e);
    return NULL;
  }
  /* swap p and q! */
  size = ntohs(encoding->sizep);
  if (size > 0) {
    rc = gcry_mpi_scan(&q,
		       GCRYMPI_FMT_USG,
		       &encoding->key[pos],
		       &size);
    pos += ntohs(encoding->sizep);
    if (rc) {
      LOG(LOG_ERROR,
	  "ERROR: could not decode hostkey (%s)\n",
	  gcry_strerror(rc));
      gcry_mpi_release(n);
      gcry_mpi_release(e);
      gcry_mpi_release(d);
      return NULL;
    }
  } else
    q = NULL;
  size = ntohs(encoding->sizeq);
  if (size > 0) {
    rc = gcry_mpi_scan(&p,
		       GCRYMPI_FMT_USG,
		       &encoding->key[pos],
		       &size);
    pos += ntohs(encoding->sizeq);
    if (rc) {
      LOG(LOG_ERROR,
	  "ERROR: could not decode hostkey (%s)\n",
	  gcry_strerror(rc));
      gcry_mpi_release(n);
      gcry_mpi_release(e);
      gcry_mpi_release(d);
      if (q != NULL)
	gcry_mpi_release(q);
      return NULL;
    }
  } else
    p = NULL;
  pos += ntohs(encoding->sizedmp1);
  pos += ntohs(encoding->sizedmq1);
  
  size = ntohs(encoding->len) - sizeof(HostKeyEncoded) - pos;
  if (size > 0) {
    rc = gcry_mpi_scan(&u,
		       GCRYMPI_FMT_USG,
		       &encoding->key[pos],
		       &size);
    if (rc) {
      LOG(LOG_ERROR,
	  "ERROR: could not decode hostkey (%s)\n",
	  gcry_strerror(rc));
      gcry_mpi_release(n);
      gcry_mpi_release(e);
      gcry_mpi_release(d);
      if (p != NULL)
	gcry_mpi_release(p);
      if (q != NULL)
	gcry_mpi_release(q);
      return NULL;
    }
  } else
    u = NULL;

  if ( (p != NULL) &&
       (q != NULL) &&
       (u != NULL) ) {
    rc = gcry_sexp_build(&res,
			 &size, /* erroff */
			 "(private-key(rsa(n %m)(e %m)(d %m)(p %m)(q %m)(u %m)))",
			 n, e, d, p, q, u);
  } else {
    if ( (p != NULL) &&
	 (q != NULL) ) {
      rc = gcry_sexp_build(&res,
			   &size, /* erroff */
			   "(private-key(rsa(n %m)(e %m)(d %m)(p %m)(q %m)))",
			   n, e, d, p, q);
    } else {
      rc = gcry_sexp_build(&res,
			   &size, /* erroff */
			   "(private-key(rsa(n %m)(e %m)(d %m)))",
			   n, e, d);
    }
  }
  gcry_mpi_release(n);
  gcry_mpi_release(e);
  gcry_mpi_release(d);
  if (p != NULL)
    gcry_mpi_release(p);
  if (q != NULL)
    gcry_mpi_release(q);
  if (u != NULL)
    gcry_mpi_release(u);

  if (rc) {
     LOG(LOG_ERROR,
	  "ERROR: could not decode hostkey (%s at %d)\n",
	  gcry_strerror(rc),
	 size);
  }
#if EXTRA_CHECKS
  if (gcry_pk_testkey(res)) {
    LOG(LOG_ERROR,
	"ERROR: decodeHostkey: hostkey is not sane!\n");
    return NULL;
  }
#endif
 return res;
#endif
}

/**
 * Encrypt a block with the public key of another host that uses the
 * same cyper.
 *
 * @param block the block to encrypt
 * @param size the size of block
 * @param publicKey the encoded public key used to encrypt
 * @param target where to store the encrypted block
 * @returns SYSERR on error, OK if ok
 **/
int encryptHostkey(void * block, 
		   unsigned short size,
		   PublicKey * publicKey,
		   RSAEncryptedData * target) {
#ifdef USE_OPENSSL
  HOSTKEY foreignkey;
  int rs;
  int len;

  foreignkey = public2Hostkey(publicKey);
  if (foreignkey == NULL)
    return SYSERR;
  rs = RSA_size(foreignkey);
  /* now encrypt. First get size of the block */
  if (size > (rs - 41)) {
    LOG(LOG_ERROR,
	"ERROR: HostKey::encryptHostkey() called with %d"
	" bytes where foreignkey allows only %d\n",
	size, 
	rs-41);
    freeHostkey(foreignkey);
    return SYSERR;
  }
  len = RSA_public_encrypt(size, 
			   block, 
			   &target->encoding[0], 
			   foreignkey,
			   RSA_PKCS1_PADDING);
  if (len != RSA_ENC_LEN) {
    LOG(LOG_ERROR,
	"ERROR: RSA-Encoding has unexpected length (%d)!",
	len);
    freeHostkey(foreignkey);
    return SYSERR;
  }
  freeHostkey(foreignkey);
  return OK;
#else
  /* libgcrypt */
  GcrySexp result;
  GcrySexp data;
  GcrySexp pubkey;
  GcryMPI val;
  GcryMPI rval;
  size_t isize;
  size_t erroff;
  int rc;
  
  pubkey = public2Hostkey(publicKey);
  isize = size;
  rc = gcry_mpi_scan(&val,
		     GCRYMPI_FMT_USG,
		     block,
		     &isize);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: encryptHostkey - gcry_mpi_scan failed (%s)\n", 
	gcry_strerror (rc));
    freeHostkey(pubkey);
    return SYSERR;
  }
  rc = gcry_sexp_build(&data,
		       &erroff,
		       "(data (flags pkcs1)(value %m))",
		       val);
  gcry_mpi_release(val);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: encryptHostkey - gcry_sexp_build failed (%s at offset %d)\n", 
	gcry_strerror (rc),
	erroff);
    freeHostkey(pubkey);
    return SYSERR;
  }
  
  rc = gcry_pk_encrypt(&result, data, pubkey);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: encryptHostkey - gcry_pk_encrypt failed (%s)\n", 
	gcry_strerror (rc),
	erroff);
    gcry_sexp_release(data);
    freeHostkey(pubkey);
    return SYSERR;
  }
  gcry_sexp_release(data);
  freeHostkey(pubkey);

  rc = key_from_sexp(&rval,
		     result,
		     "rsa",
		     "a");
  gcry_sexp_release(result);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: hostkeyEncrypt: key_from_sexp failed (%d, %s)\n",
	rc, 
	gcry_strerror(rc));
    return SYSERR;
  }
  isize = sizeof(RSAEncryptedData);
  rc = gcry_mpi_print(GCRYMPI_FMT_USG,
		      (char*)target,
		      &isize, 
		      rval);
  gcry_mpi_release(rval);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: encryptHostkey - gcry_mpi_print failed (%s)\n", 
	gcry_strerror (rc));
    return SYSERR;
  }
  return OK;
#endif
}

/**
 * Decrypt a given block with the hostkey. 
 *
 * @param hostkey the hostkey with which to decrypt this block
 * @param block the data to decrypt, encoded as returned by encrypt
 * @param result pointer to a location where the result can be stored
 * @param max the maximum number of bits to store for the result, if
 *        the decrypted block is bigger, an error is returned
 * @returns the size of the decrypted block, -1 on error
 **/
int decryptHostkey(HOSTKEY hostkey, 
		   RSAEncryptedData * block,
		   void * result,
		   unsigned int max) {
#ifdef USE_OPENSSL
  RSAEncryptedData tmp; /* this is as big as the result can possibly get */
  int size;

  if (block == NULL)
    return -1;

  size = RSA_private_decrypt(sizeof(RSAEncryptedData), 
			     &block->encoding[0],
			     &tmp.encoding[0], 
			     hostkey,
			     RSA_PKCS1_PADDING);
  if ( (size == -1) || 
       (size > max) )
    return -1;
  memcpy(result,
	 &tmp.encoding[0],
	 size);
  return size;
#else
  GcrySexp resultsexp;
  GcrySexp data;
  size_t erroff;
  size_t size;
  GcryMPI val;
  int rc;
  char * endp;
  char * tmp;

#if EXTRA_CHECKS
  if (gcry_pk_testkey(hostkey)) {
    LOG(LOG_ERROR,
	"ERROR: decrypt: hostkey is not sane!\n");
    return -1;
  }
#endif
  size = sizeof(RSAEncryptedData);
  rc = gcry_mpi_scan(&val,
		     GCRYMPI_FMT_USG,
		     &block->encoding[0],
		     &size);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: hostkeyDecrypt: gcry_mpi_scan failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }
  rc = gcry_sexp_build(&data,
		       &erroff,
		       "(enc-val(flags)(rsa(a %m)))",
		       val);
  gcry_mpi_release(val);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: hostkeyDecrypt: gcry_sexp_build failed (%s at %d)\n",
	gcry_strerror(rc),
	erroff);
    return SYSERR;
  }
  rc = gcry_pk_decrypt(&resultsexp,
		       data,
		       hostkey);
  gcry_sexp_release(data);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: hostkeyDecrypt: gcry_pk_decrypt failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }

  /* resultsexp has format "(value %m)" */
  val = gcry_sexp_nth_mpi(resultsexp, 1, GCRYMPI_FMT_USG);
  gcry_sexp_release(resultsexp);
  if (val == NULL) {
    LOG(LOG_ERROR,
	"ERROR: gcry_sexp_nth_mpi failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }
  tmp = xmalloc(max + 500,
		"allocating buffer for pkcs1 decoding");
  size = max+500;
  rc = gcry_mpi_print(GCRYMPI_FMT_USG,
		      tmp,
		      &size,
		      val);
  gcry_mpi_release(val);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: gcry_mpi_print failed (%s)\n",
	gcry_strerror(rc));
    xfree(tmp,
	  "freeing pcks block");
    return SYSERR;
  }

  endp = tmp;
  if (*endp == 0) {
    endp++;
    size--;
  }
  if ( (!size) || 
       (*endp != 0x02 )) {
    LOG(LOG_ERROR,
	"ERROR: not a pkcs-1 block type 2 (size=%d, *tmp=%d)!\n",
	size, (int)*tmp);
    xfree(tmp,
	  "freeing pcks block");
    return SYSERR;
  }
  /* serach the end of the padding */
  while ( (size > 0) && 
	  ((*endp) != 0) ) {
    size--;
    endp++;
  }
  if ( (size == 0) || 
       (*endp != 0x0) ) {
    LOG(LOG_ERROR,
	"ERROR: not a pkcs-1 block type 2 (size=%d, *endp=%d)\n",
	size, (int) *endp);
    xfree(tmp,
	  "freeing pcks block");
    return SYSERR;
  }
  size--;
  endp++;

  if (size > max)
    size = max;
  memcpy(result,
	 endp,
	 size);
  xfree(tmp,
	"freeing pcks block");
  return size;
#endif
}

/**
 * Sign a given block.
 *
 * @param hostkey the hostkey with which to sign this block
 * @param size how many bytes to sign
 * @param block the data to sign
 * @param result where to write the signature
 * @return SYSERR on error, OK on success
 **/
int sign(HOSTKEY hostkey, 
	 unsigned short size,
	 void * block,
	 Signature * sig) {
#ifdef USE_OPENSSL
#if EXTRA_CHECKS
  PublicKey pkey;
#endif
  int rs = RSA_size(hostkey);
  int sigSize;
  HashCode160 hc;

  if (block == NULL)
    return SYSERR;
  if (rs != sizeof(Signature)) {
    LOG(LOG_ERROR,
	"ERROR: sign: signature length (RSA_size) has unexpected value (%d)!",
	rs);
    return SYSERR;
  }
  hash(block, 
       size,
       &hc);
  if (1 != RSA_sign(NID_ripemd160,
		    (char*)&hc,
		    sizeof(HashCode160),
		    &sig->sig[0],
		    &sigSize,
		    hostkey)) {
    LOG(LOG_ERROR,
	"ERROR: RSA_sign failed!",
	sigSize);
    return SYSERR;
  }
  if (sigSize != sizeof(Signature)) {
    LOG(LOG_ERROR,
	"ERROR: sign: sigSize wrong (%u)!",
	sigSize);
    return SYSERR;
  }
#if EXTRA_CHECKS
  getPublicKey(hostkey, &pkey);
  if (SYSERR == verifySig(block, size, sig, &pkey)) {
    errexit("FATAL: sign: generated signature does not pass verification!\n");
  }
#endif
  return OK;
#else
  /* libgcrypt */
  GcrySexp result;
  GcrySexp data;
  size_t ssize;
  GcryMPI rval;
  HashCode160 hc;
  char * buff;
  int bufSize;
  int rc;

  hash(block, size, &hc);
#define FORMATSTRING "(4:data(5:flags5:pkcs1)(4:hash6:rmd16020:01234567890123456789))"
  bufSize = strlen(FORMATSTRING) + 1;
  buff = xmalloc(bufSize,
		 "sexp building");
  memcpy(buff,
	 FORMATSTRING,
	 bufSize);
  memcpy(&buff[bufSize - strlen("012345678901234567890))")],
	 &hc,
	 sizeof(HashCode160));
  rc = gcry_sexp_new(&data,
		     buff,
		     bufSize, 
		     0);
  xfree(buff,
	"sexp building buffer");
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: sign: gcry_sexp_new failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }
  rc = gcry_pk_sign(&result, data, hostkey);
  gcry_sexp_release(data);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: sign: gcry_pk_sign failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }
  rc = key_from_sexp(&rval,
		     result,
		     "rsa", 
		     "s");
  gcry_sexp_release(result);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: sign: key_from_sexp failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }
  ssize = sizeof(Signature);
  rc = gcry_mpi_print(GCRYMPI_FMT_USG,
		      (char*)sig,
		      &ssize,
		      rval);
  gcry_mpi_release(rval);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: sign: gcry_mpi_print failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }
  return OK;
#endif
}

/**
 * Verify signature.
 *
 * @param block the signed data
 * @param len the length of the block 
 * @param sig signature
 * @param publicKey public key of the signer
 * @returns OK if ok, SYSERR if invalid
 **/
int verifySig(void * block,
	      unsigned short len,
	      Signature * sig,	      
	      PublicKey * publicKey) {
#ifdef USE_OPENSSL
  HOSTKEY hostkey;
  int rs;
  HashCode160 hc;
 
  hostkey = public2Hostkey(publicKey);
  if ((hostkey == NULL) || (sig == NULL) || (block == NULL))
    return SYSERR; /* hey, no data !? */
  rs = RSA_size(hostkey);
  if (rs != RSA_ENC_LEN) {
    errexit("FATAL: verifySig: rs != RSA_ENC_LEN (%d)!",
	    rs);
  }
  hash(block, len, &hc);
  if (1 != RSA_verify(NID_ripemd160,
		      &hc,
		      sizeof(HashCode160),
		      &sig->sig[0],
		      sizeof(Signature),
		      hostkey)) {
    LOG(LOG_ERROR,
	"ERROR: verifySig: signature mismatch\n");
    freeHostkey(hostkey);
    return SYSERR;
  }
  freeHostkey(hostkey);
  return OK;
#else
  /* libgcrypt */
  GcrySexp data;
  GcrySexp sigdata;
  size_t size;
  GcryMPI val;
  HOSTKEY hostkey;
  HashCode160 hc;
  char * buff;
  int bufSize;
  size_t erroff;
  int rc;
 
  size = sizeof(Signature);
  rc = gcry_mpi_scan(&val,
		     GCRYMPI_FMT_USG,
		     (char*)sig,
		     &size);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: sign: key_from_sexp failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }
  rc = gcry_sexp_build(&sigdata,
		       &erroff,
		       "(sig-val(rsa(s %m)))",
		       val);
  gcry_mpi_release(val);
  if (rc) {
    LOG(LOG_ERROR,
	"ERROR: sign: gcry_sexp_build failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  }  
  hash(block, len, &hc);
  bufSize = strlen(FORMATSTRING) + 1;
  buff = xmalloc(bufSize,
		 "sexp building");
  memcpy(buff,
	 FORMATSTRING,
	 bufSize);
  memcpy(&buff[strlen(FORMATSTRING) - strlen("01234567890123456789))")],
	 &hc,
	 sizeof(HashCode160));
  rc = gcry_sexp_new(&data,
		     buff,
		     bufSize, 
		     0);
  xfree(buff,
	"sexp building");
  hostkey = public2Hostkey(publicKey);
  rc = gcry_pk_verify(sigdata,
		      data,
		      hostkey);  
  freeHostkey(hostkey);
  gcry_sexp_release(data);
  gcry_sexp_release(sigdata);
  if (rc) {
    LOG(LOG_WARNING,
	"WARNING: signature verification failed (%s)\n",
	gcry_strerror(rc));
    return SYSERR;
  } else
    return OK;
#endif
}


/* end of hostkey.c */
